@@ -16,7 +16,7 @@ from openevolve.database import Program, ProgramDatabase
 from openevolve.evaluator import Evaluator
 from openevolve.llm.ensemble import LLMEnsemble
 from openevolve.prompt.sampler import PromptSampler
-from openevolve.threaded_parallel import ImprovedParallelController
+from openevolve.process_parallel import ProcessParallelController
 from openevolve.utils.code_utils import (
     extract_code_language,
 )
@@ -157,6 +157,7 @@ class OpenEvolve:
             self.llm_evaluator_ensemble,
             self.evaluator_prompt_sampler,
             database=self.database,
+            suffix=Path(self.initial_program_path).suffix,
         )
         self.evaluation_file = evaluation_file
 
@@ -249,6 +250,23 @@ class OpenEvolve:
             )
 
             self.database.add(initial_program)
+
+            # Check if combined_score is present in the metrics
+            if "combined_score" not in initial_metrics:
+                # Calculate average of numeric metrics
+                numeric_metrics = [
+                    v
+                    for v in initial_metrics.values()
+                    if isinstance(v, (int, float)) and not isinstance(v, bool)
+                ]
+                if numeric_metrics:
+                    avg_score = sum(numeric_metrics) / len(numeric_metrics)
+                    logger.warning(
+                        f"⚠️  No 'combined_score' metric found in evaluation results. "
+                        f"Using average of all numeric metrics ({avg_score:.4f}) for evolution guidance. "
+                        f"For better evolution results, please modify your evaluator to return a 'combined_score' "
+                        f"metric that properly weights different aspects of program performance."
+                    )
         else:
             logger.info(
                 f"Skipping initial program addition (resuming from iteration {start_iteration} "
@@ -257,7 +275,7 @@ class OpenEvolve:
 
         # Initialize improved parallel processing
         try:
-            self.parallel_controller = ImprovedParallelController(
+            self.parallel_controller = ProcessParallelController(
                 self.config, self.evaluation_file, self.database
             )
 
@@ -280,9 +298,20 @@ class OpenEvolve:
 
             self.parallel_controller.start()
 
+            # When starting from iteration 0, we've already done the initial program evaluation
+            # So we need to adjust the start_iteration for the actual evolution
+            evolution_start = start_iteration
+            evolution_iterations = max_iterations
+
+            # If we just added the initial program at iteration 0, start evolution from iteration 1
+            if should_add_initial and start_iteration == 0:
+                evolution_start = 1
+                # User expects max_iterations evolutionary iterations AFTER the initial program
+                # So we don't need to reduce evolution_iterations
+
             # Run evolution with improved parallel processing and checkpoint callback
             await self._run_evolution_with_checkpoints(
-                start_iteration, max_iterations, target_score
+                evolution_start, evolution_iterations, target_score
             )
 
         finally:
@@ -324,10 +353,20 @@ class OpenEvolve:
                     best_program = best_by_combined
 
         if best_program:
-            logger.info(
-                f"Evolution complete. Best program has metrics: "
-                f"{format_metrics_safe(best_program.metrics)}"
-            )
+            if (
+                hasattr(self, "parallel_controller")
+                and self.parallel_controller
+                and self.parallel_controller.early_stopping_triggered
+            ):
+                logger.info(
+                    f"🛑 Evolution complete via early stopping. Best program has metrics: "
+                    f"{format_metrics_safe(best_program.metrics)}"
+                )
+            else:
+                logger.info(
+                    f"Evolution complete. Best program has metrics: "
+                    f"{format_metrics_safe(best_program.metrics)}"
+                )
             self._save_best_program(best_program)
             return best_program
         else:
@@ -438,12 +477,17 @@ class OpenEvolve:
             start_iteration, max_iterations, target_score, checkpoint_callback=self._save_checkpoint
         )
 
-        # Check if shutdown was requested
-        if self.parallel_controller.shutdown_flag.is_set():
+        # Check if shutdown or early stopping was triggered
+        if self.parallel_controller.shutdown_event.is_set():
             logger.info("Evolution stopped due to shutdown request")
             return
+        elif self.parallel_controller.early_stopping_triggered:
+            logger.info("Evolution stopped due to early stopping - saving final checkpoint")
+            # Continue to save final checkpoint for early stopping
 
         # Save final checkpoint if needed
+        # Note: start_iteration here is the evolution start (1 for fresh start, not 0)
+        # max_iterations is the number of evolution iterations to run
         final_iteration = start_iteration + max_iterations - 1
         if final_iteration > 0 and final_iteration % self.config.checkpoint_interval == 0:
             self._save_checkpoint(final_iteration)
